// ***********************************************************/
// *  scanner.flex                                           */
// *  Procesadores de Lenguaje-I CURSO 2014/2015             */
// *  Diseo e implementacin de un compilador de cUNED      */
// *  Grupo especificacion: B                                */
// *                                                         */
// *  Autor: Rafael Lacalle Merino                           */
// *  Valencia, 15 de enero de 2015                          */
// ***********************************************************/

// Importaciones

package compiler.lexical;

import compiler.syntax.sym;
import compiler.lexical.Token;
import es.uned.lsi.compiler.lexical.ScannerIF;
import es.uned.lsi.compiler.lexical.LexicalError;
import es.uned.lsi.compiler.lexical.LexicalErrorManager;

// incluir aqui, si es necesario otras importaciones


// Directivas JLEX, macros, configuracin de comportamientos

%%
 
%public
%class Scanner
%char
%line
%column
%cup

%implements ScannerIF
%scanerror LexicalError

// incluir aqui, si es necesario otras directivas
%full   //directiva que utiliza el codigo ASCII extendido

%state COMENTARIOS

%{
  LexicalErrorManager lexicalErrorManager = new LexicalErrorManager ();
  private int commentCount = 0;
%} 

%eof{
	if (commentCount>0) {
							// error en caso de comentarios mal balanceados
							LexicalError error = new LexicalError();
							error.setLine (yyline + 1);
							error.setColumn (yycolumn + 1);
							error.setLexema (yytext ());
							lexicalErrorManager.lexicalError ("Comentarios mal balanceados, falta por cerrar " + commentCount + " comentarios. Error: " + error);							
						}
%eof}
  
DIGITO=[0-9]      // definicion digitos
LETRA=[a-zA-Z]    // definicion letras

NUMERO={DIGITO}+  // definicion de numero
IDENTIFICADOR=({LETRA}({LETRA}|{DIGITO})*)  // definicion de identificador

ESPACIO_BLANCO=([\ \t\b\r\n\f])+  // espacio en blanco
COMENTARIO_ABRIR=("/*")           // abrir comentario
COMENTARIO_CERRAR=("*/")          // cerrar comentario
CONSTANTE_CADENA=[^\"]*           // cadena
fin = "fin"{ESPACIO_BLANCO}       // fin

%%


// ----------------
//  Estado Inicial
// ----------------

<YYINITIAL> 
{

// --------------------------
//   Gestin de comentarios
// --------------------------
	// Si en el estado inicial encontramos COMENTARIO_CERRAR sin su apertura correspondiente, mostramos un error
	// y cambiamos al estado COMENTARIOS para gestionar el posible comentario sin generar tokens vlidos.
	{COMENTARIO_CERRAR}	{
							LexicalError error = new LexicalError();
							error.setLine (yyline + 1);
							error.setColumn (yycolumn + 1);
							error. setLexema (yytext ());
							lexicalErrorManager.lexicalError ("Se ha encontrado un cierre de comentario ["
							+ (yyline + 1) + ";" + (yycolumn + 1) + "], sin existir comentarios abiertos. " + error);		
						}
    // Si en el estado inicial encontramos COMENTARIO_ABRIR, cambiamos al estado COMENTARIOS
    // para gestionar los comentarios sin tenerlos en cuenta como tokens.
	{COMENTARIO_ABRIR}	{
							commentCount = commentCount + 1;
							yybegin(COMENTARIOS);							
						}

// -----------------------
//   Palabras reservadas
// -----------------------

   "#define"            {  
                           Token token = new Token (sym.BLOQUE_CONSTANTES);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

 	"if"	            {  
                           Token token = new Token (sym.SI);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
   "else"               {  
                           Token token = new Token (sym.SINO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
 
    "int"	            {  
                           Token token = new Token (sym.ENTERO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

    "main"	            {  
                           Token token = new Token (sym.PRINCIPAL);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "printc"            {  
                           Token token = new Token (sym.IMPRIME_CADENA);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "printi"            {  
                           Token token = new Token (sym.IMPRIME_ENTERO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "return"            {  
                           Token token = new Token (sym.RETORNO_FUNCION);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "void"              {  
                           Token token = new Token (sym.TIPO_VACIO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "while"             {  
                           Token token = new Token (sym.BUCLE_WHILE);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                                                                                               
// -----------------------
//   Delimitadores
// -----------------------

   "("                {  
                           Token token = new Token (sym.PARENT_ABRIR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }                        

   ")"                {  
                           Token token = new Token (sym.PARENT_CERRAR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
   "["                {  
                           Token token = new Token (sym.MATRIZ_ABRIR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
   "]"                {  
                           Token token = new Token (sym.MATRIZ_CERRAR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }                                                

   "{"                {  
                           Token token = new Token (sym.BLOQUE_CODIGO_ABRIR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
   "}"                {  
                           Token token = new Token (sym.BLOQUE_CODIGO_CERRAR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

   ","                {  
                           Token token = new Token (sym.COMA);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

   ";"                {  
                           Token token = new Token (sym.FIN_SENTENCIA);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

   "&"                {  
                           Token token = new Token (sym.PASO_REFERENCIA);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }


// --------------------------
//   Operadores aritmticos
// --------------------------

           			       
    "+"                {  
                           Token token = new Token (sym.MAS);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
// ---------------------------
//   Operadores relacionales
// ---------------------------

           			       
    "<"                {  
                           Token token = new Token (sym.MENOR_QUE);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
    "!="                {  
                           Token token = new Token (sym.DISTINTO_QUE);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }

// -----------------------
//   Operadores lgicos
// -----------------------

           			       
    "&&"                {  
                           Token token = new Token (sym.AND_LOGICO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                        
// ----------------------------
//   Operadores de asignacin
// ----------------------------

           			       
    "="                {  
                           Token token = new Token (sym.ASIGNACION);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;
                        }
                         
// ----------------------------
//   Operadores de acceso
// ----------------------------
// El nico operador de acceso . -> acceso al campo de un registro, no es aplicable al grupo B.                             
       
    // incluir aqui el resto de las reglas patron - accion

// ----------------------------
//   Patrones
// ----------------------------

	{ESPACIO_BLANCO} {}	
	
	{NUMERO} 			{
						   Token token = new Token (sym.NUMERO);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;	
						}												

						
	{IDENTIFICADOR} 	{
						   Token token = new Token (sym.IDENTIFICADOR);
                           token.setLine (yyline + 1);
                           token.setColumn (yycolumn + 1);
                           token.setLexema (yytext ());
           			       return token;	
						}
						
    \"{CONSTANTE_CADENA}\" {
							   Token token = new Token (sym.CONSTANTE_CADENA);
	                           token.setLine (yyline + 1);
	                           token.setColumn (yycolumn + 1);
	                           token.setLexema (yytext ());
	           			       return token;		
						   }
	{fin} {}
    
    // error en caso de no coincidir con ningn patrn
	[^]     
                        {                                               
                           LexicalError error = new LexicalError ();
                           error.setLine (yyline + 1);
                           error.setColumn (yycolumn + 1);
                           error.setLexema (yytext ());
                           lexicalErrorManager.lexicalError (error);                           
                        }
    
} // Fin del estado YYINITIAL


// ---------------------
//  Estado Comentarios
// ---------------------

<COMENTARIOS>
{
	// Si en el estado COMENTARIOS aparece un nuevo comentario, incrementamos el contador
	{COMENTARIO_ABRIR}	{
							commentCount = commentCount + 1;
						}
	// Si encontramos el cierre de un comentario, decrementamos el contador y comprobamos su finalizacin
	{COMENTARIO_CERRAR}	{
							commentCount = commentCount - 1;
							if (commentCount==0)
							{
								yybegin(YYINITIAL);
							}
						}
	// cualquier otra cosa es ignorada
	[^]   {}
						
} // Fin del estado COMENTARIOS
